/*
 * linux/drivers/sound/pcsp.c
 *
 * /dev/pcsp implementation
 *
 * Copyright (C) 1993-1997  Michael Beck 
 */

#include <linux/config.h>
#include <linux/soundcard.h>
#include <linux/pcsp.h>

#include <linux/fs.h>
#include <linux/errno.h>
#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/init.h>

#ifdef MODULE
#include <linux/module.h>
#include <linux/version.h>
#include <linux/sound.h>

MODULE_AUTHOR("Michael Beck <beck@DResearch.de>");
MODULE_DESCRIPTION("Alternate Sound Driver for Linux 2.x Version 1.3");
MODULE_SUPPORTED_DEVICE("pcsp");
EXPORT_NO_SYMBOLS;
#endif

/* 16bit emulation needs the mixer */
#ifdef CONFIG_PCSP_16BIT
#define CONFIG_PCSP_MIXER
#endif

extern struct file_operations pcsp_dsp_fops, pcsp_mixer_fops;
extern long pcsp_driver_init(long);
extern int pcsp_set_format(int);
extern void pcsp_free_ports(void);


#ifdef MODULE

static int pcsp_sound_dsp, pcsp_sound_mixer;

__initfunc(int init_module(void))
{
	int minor, major;

        if (! pcsp_enabled) {
                printk("pcsp disabled\n");
                return -ENODEV;
        }
	
	if ((pcsp_sound_dsp = register_sound_dsp(&pcsp_dsp_fops, -1))< 0) {
		printk(KERN_WARNING "Unable to register PC speaker DSP device\n");
		return pcsp_sound_dsp;
	}
	if ((pcsp_sound_mixer = register_sound_mixer (&pcsp_mixer_fops, -1))<0) {
		printk(KERN_WARNING "Unable to register PC speaker mixer device\n");
		unregister_sound_dsp(pcsp_sound_dsp);
		return pcsp_sound_mixer;
	}
	
	printk(KERN_DEBUG "PCSP on device %d, mixer on device %d\n",pcsp_sound_dsp, pcsp_sound_mixer);
	
	major = (PCSP_SOUND_VERSION >> 8);
	minor = (PCSP_SOUND_VERSION & 0xFF);
	if (! (minor & 0xF))
		minor >>= 4;

	pcsp_device_init();
#ifdef CONFIG_PCSP_MIXER
	pcsp_mixer_init();
#endif
	return 0;
}

void cleanup_module(void)
{
	if (MOD_IN_USE)
		printk("pcsnd: busy - remove delayed\n");
	else {
		pcsp_free_ports();
		unregister_sound_dsp(pcsp_sound_dsp);
		unregister_sound_mixer (pcsp_sound_mixer);
	}
}
#endif
