/*
 * linux/drivers/sound/pcsp_mixer.c
 * 
 * /dev/pcsp implementation - simple Mixer routines
 * 
 * (C) 1993-1997  Michael Beck
 * Craig Metz (cmetz@thor.tjhsst.edu)
 */

#include <linux/config.h>

#if defined(CONFIG_PCSP_16BIT) || defined(CONFIG_PCSP_MIXER)

#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/errno.h>
#include <linux/soundcard.h>
#include <linux/pcsp.h>
#include <asm/uaccess.h>

#ifdef MODULE
#define __NO_VERSION__
#include <linux/module.h>
#include <linux/version.h>
#else
#define MOD_INC_USE_COUNT
#define MOD_DEC_USE_COUNT
#endif

#define IOCTL_OUT(arg, ret)             (put_user(ret, (long *)arg))
#define OFF				0

static int      rec_devices = 0;	/* No recording source */

static struct {
	char		active;
} pcmixer;

#define POSSIBLE_RECORDING_DEVICES	(0)
#define SUPPORTED_MIXER_DEVICES		(SOUND_MASK_VOLUME)

static unsigned short levels[SOUND_MIXER_NRDEVICES] =
{
	0,				/* Master Volume */
	/* The next devices are not supported, so they are zero */
	0,				/* Bass */
	0,				/* Treble */
	0,				/* FM */
	0,				/* PCM */
	0,				/* PC Speaker */
	0,				/* Ext Line */
	0,				/* Mic */
	0,				/* CD */
	0,				/* Recording monitor */
	0,				/* SB PCM */
	0				/* Recording level */
};

extern void pcsp_set_volume(unsigned short);
extern unsigned pcsp_get_mode(void);

int pcsp_mixer_set(int whichDev, unsigned int level)
{
	int left, right;

	left = level & 0x7f;
	right = (level & 0x7f00) >> 8;

	switch (whichDev) {
	case SOUND_MIXER_VOLUME:	/* Master volume (0-127) */
		levels[whichDev] = left | (right << 8);
		pcsp_set_volume(levels[whichDev]);
		break;

	default:
		return (-EINVAL);
	}
	return (levels[whichDev]);
}

int pcsp_mixer_ioctl(struct inode * inode, struct file * file,
                        unsigned int cmd, unsigned long arg)
{
	int val;
	long *ptr = (long *)arg;

	if (((cmd >> 8) & 0xff) == 'M') {
		if (cmd & IOC_IN) {
			get_user_ret(val, ptr, -EINVAL);
			if ((cmd & 0xff)< SOUND_MIXER_NRDEVICES)
				return IOCTL_OUT(arg, pcsp_mixer_set(cmd & 0xff, val));
			else {
				return (-EINVAL); 
			}
		}
		
		switch (cmd & 0xff) {
			
		case SOUND_MIXER_RECSRC:
			return IOCTL_OUT(arg, rec_devices);
			
		case SOUND_MIXER_STEREODEVS:
			return IOCTL_OUT(arg, SUPPORTED_MIXER_DEVICES & ~(SOUND_MASK_BASS | SOUND_MASK_TREBLE));
			
		case SOUND_MIXER_DEVMASK:
			return IOCTL_OUT(arg, SUPPORTED_MIXER_DEVICES);
			
		case SOUND_MIXER_RECMASK:
			return IOCTL_OUT(arg, POSSIBLE_RECORDING_DEVICES & SUPPORTED_MIXER_DEVICES);
			
		case SOUND_MIXER_CAPS:
			return IOCTL_OUT(arg, 0);
			
		default:
			if ((cmd & 0xff) < SOUND_MIXER_NRDEVICES)
				return IOCTL_OUT(arg, levels[cmd & 0xff]);
			else {
				return (-EINVAL);
			}
		}
	}

	return (-EINVAL);
}

static int pcsp_mixer_release(struct inode * inode, struct file * file)
{
	--pcmixer.active;
	MOD_DEC_USE_COUNT;
	return 0;
}

static int pcsp_mixer_open(struct inode * inode, struct file * file)
{
	if (pcmixer.active)
		return (-EBUSY);
	++pcmixer.active;
	MOD_INC_USE_COUNT;
	return 0;
}

struct file_operations pcsp_mixer_fops = {
        NULL,           /* pcsp_mixer_seek */
        NULL,		/* pcsp_mixer_read */
        NULL,		/* pcsp_mixer_write */
        NULL,           /* pcsp_readdir */
        NULL,           /* pcsp_select */
        pcsp_mixer_ioctl,
        NULL,           /* pcsp_mmap */
        pcsp_mixer_open,
	NULL,
        pcsp_mixer_release,
};

int pcsp_mixer_init(void)
{
	levels[0] = ((pcsp.right*100 + 128) & ~0xFF) | ((pcsp.left*100 + 128) >> 8);
	pcsp_mixer_set(SOUND_MIXER_VOLUME, levels[SOUND_MIXER_VOLUME]);
	pcmixer.active = 0;
	return 0;
}

#endif
